#!/usr/bin/perl
use warnings;
use strict;
use Test::More tests => 39;
use Test::Exception;

use Narada::Lock qw( shared_lock exclusive_lock unlock_new unlock child_inherit_lock );

use Time::HiRes qw( time sleep );
use Fcntl qw(:DEFAULT :flock F_SETFD FD_CLOEXEC);
use Errno;

use constant LOCKNEW    => Narada::Lock::LOCKNEW;
use constant LOCKFILE   => Narada::Lock::LOCKFILE;

use File::Temp qw( tempdir );
chomp(my $cwd=`pwd`);
$ENV{PATH} = "$cwd/blib/script:$ENV{PATH}";
$ENV{PERL5LIB} = "$cwd/blib:$ENV{PERL5LIB}";
chdir tempdir( CLEANUP => 1 )
    and system('narada-new') == 0
    or die "Unable to create project: $!";

sysopen my $F_lock, LOCKFILE, O_RDWR|O_CREAT or die "open: $!";


sub has_sh {
    my $has_sh = !flock $F_lock, LOCK_EX|LOCK_NB;
    flock $F_lock, LOCK_UN;
    return $has_sh;
}

sub has_ex {
    my $has_ex = !flock $F_lock, LOCK_SH|LOCK_NB;
    flock $F_lock, LOCK_UN;
    return $has_ex;
}

sub between {
    my ($min, $max, $code) = @_;
    my $t = time();
    $code->();
    $t = time()-$t;
    ok $min <= $t && $t <= $max, "... done in $min .. $max sec";
    return;
}


# unlock() - не выдаёт ошибок если нет блокировки
# unlock_new() - не выдаёт ошибок если нет блокировки
lives_ok { unlock()     } 'unlock without lock';
lives_ok { unlock_new() } 'unlock_new without lock';

# shared_lock() - ставит сразу
# unlock() - снимает блокировку
ok !has_sh(),   'no  shared lock before shared_lock()';
between 0, 0.5, sub { ok shared_lock(), 'shared_lock() successful' };
ok has_sh(),    'has shared lock after shared_lock()';
unlock();
ok !has_sh(),   'no  shared lock after unlock()';

# shared_lock(0) - ставит сразу
# shared_lock(2) - ставит сразу
between 0, 0.5, sub { ok shared_lock(0), 'shared_lock(0) successful' };
between 0, 0.5, sub { ok shared_lock(2), 'shared_lock(2) successful' };

# shared_lock()  при фоновом `narada-lock-exclusive sleep 1`
#   - ставит через 1 секунду
unlock(); system('narada-lock-exclusive sleep 1 & sleep 0.2');
between 0.5, 1.5, sub { ok shared_lock(), 'shared_lock() successful after exclusive' };
# shared_lock(0) при фоновом `narada-lock-exclusive sleep 1`
#   - сразу возвращает ошибку
unlock(); system('narada-lock-exclusive sleep 1 & sleep 0.2');
between 0, 0.5, sub { ok !shared_lock(0), 'shared_lock(0) failed while exclusive' };
shared_lock();  # wait for narada-lock-exclusive finish
# shared_lock(2) при фоновом `narada-lock-exclusive sleep 1`
#   - ставит через 1 секунду
unlock(); system('narada-lock-exclusive sleep 1 & sleep 0.2');
between 0.5, 1.5, sub { ok shared_lock(2), 'shared_lock(2) successful after exclusive' };
# shared_lock(1) при фоновом `narada-lock-exclusive sleep 2`
#   - возвращает ошибку через 1 секунды
unlock(); system('narada-lock-exclusive sleep 2 & sleep 0.2');
between 0.5, 1.5, sub { ok !shared_lock(1), 'shared_lock(1) failed while exclusive' };
shared_lock();  # wait for narada-lock-exclusive finish
unlock();

# exclusive_lock() - ставит сразу, создаёт файл LOCKNEW
# unlock_new() - удаляет файл LOCKNEW, блокировка остаётся
# unlock() - снимает блокировку
ok !has_ex(),   'no  exclusive lock before exclusive_lock()';
ok !-e LOCKNEW, 'no  LOCKNEW file   before exclusive_lock()';
between 0, 0.5, sub { exclusive_lock() };
ok has_ex(),    'has exclusive lock after exclusive_lock()';
ok -e LOCKNEW,  'has LOCKNEW file   after exclusive_lock()';
unlock_new();
ok has_ex(),    'has exclusive lock after unlock_new()';
ok !-e LOCKNEW, 'no  LOCKNEW file   after unlock_new()';
unlock();
ok !has_ex(),   'no  exclusive lock after unlock()';

# exclusive_lock() при фоновом `narada-lock sleep 1`
#   - ставит через 1 секунду
unlock_new(); unlock(); system('narada-lock sleep 1 & sleep 0.2');
between 0.5, 1.5, sub { exclusive_lock() };
# exclusive_lock() при фоновом `narada-lock-exclusive sleep 1`
#   - ставит через 1 секунду, создаёт файл LOCKNEW
unlock_new(); unlock(); system('narada-lock-exclusive sleep 1 & sleep 0.2');
between 0.5, 1.5, sub { exclusive_lock() };
ok -e LOCKNEW, 'has LOCKNEW file after exclusive_lock()';
# exclusive_lock(), unlock()
#   * shared_lock(0)
#       - сразу возвращает ошибку
#   * unlock_new(), shared_lock(0)
#       - ставит сразу
unlock_new(); unlock();
exclusive_lock();
unlock();
between 0, 0.5, sub { ok !shared_lock(0), 'shared_lock(0) failed with LOCKNEW' };
unlock_new();
between 0, 0.5, sub { ok shared_lock(0), 'shared_lock(0) successful without LOCKNEW' };

# запустить процесс, который сделает shared_lock(), запустит sleep 1 & и выйдет
#   - при этом блокировки быть не должно
unlock_new(); unlock(); system('perl','-MNarada::Lock=shared_lock,child_inherit_lock','-e',
    'shared_lock(); system("sleep 1 &")');
ok !has_sh(), 'no  shared lock without child_inherit_lock(1)';
# запустить процесс, который сделает shared_lock(), child_inherit_lock(1),
# запустит sleep 1 & sleep 2 & и выйдет
#   - при этом блокировка должна быть, и должна сняться через 2 секунды
unlock_new(); unlock(); system('perl','-MNarada::Lock=shared_lock,child_inherit_lock','-e',
    'shared_lock(); child_inherit_lock(1); system("sleep 1 & sleep 2 &")');
ok has_sh(),  'has shared lock with child_inherit_lock(1)';
sleep 2.5;
ok !has_sh(), 'no  shared lock after childs exit';
# запустить процесс, который сделает shared_lock(), child_inherit_lock(1),
# запустит sleep 1 &, сделает child_inherit_lock(0), запустит sleep 2 & и выйдет
#   - при этом блокировка должна быть, и должна сняться через 1 секунду
unlock_new(); unlock(); system('perl','-MNarada::Lock=shared_lock,child_inherit_lock','-e',
    'shared_lock(); child_inherit_lock(1); system("sleep 1 &");
     child_inherit_lock(0); system("sleep 2 &")');
ok has_sh(),  'has shared lock with child_inherit_lock(1)';
sleep 1.5;
ok !has_sh(), 'no  shared lock after first child exit';

chdir '/';  # work around warnings in File::Temp CLEANUP handler
